# Sid Touati and Julien Worms, 2017
# This is a demo for VARCORE

source("VARCORE.R")

# Demo1

# Load a sample of data from a file, named C1, execution times in seconds
C1 <- read.csv("ammp-C2", header=F)$V1

# Clustering: building a Gaussian mixture model for the dataset C1
# Be careful, data inside a sample must not be identical otherwise no clustering is possible
clusC1 <- VARCORE_Clustering(C1)

# Plot the figure
VARCORE_plot_clustering_result(clusC1,C1, maintitle="ammp", subtitle="Config 1", plot_legend=T, plot_rug=T, xlabel="Execution Times", ylabel="Probability")
VARCORE_plot_clustering_result(clusC1,C1) # it also works, but with less details

# Extract cluster information and print them
clsinfo <- VARCORE_extract_cluster_information(clusC1)
clsinfo
# Every line corresponds to a cluster (Gaussian) with its weight, mean and variance
# Here we have 5 clusters
#weights    means      stdevs
#1 0.09677359 92.21333 0.163372026
#2 0.15280620 93.26964 0.146702827
#3 0.45830339 93.54552 0.227167773
#4 0.16059283 94.21802 0.003986046
#5 0.13152400 94.99607 0.392762363

# Let us compute the variability level of C1:
nbmodes = VARCORE_nbmodes_estimation(clusC1)
nbmodes
#[1] 4
# So we say that the variability level of C1 is equal to 4, even if its number of  clusters is equal to 5.

# Extract the data that belong to every cluster (here we have 5 clusters)
VARCORE_extract_cluster_list(clusC1,C1)
#[[1]]
#[1] 92.41 92.01 92.22

#[[2]]
#[1] 93.22 93.21 93.21 93.02 93.21

#[[3]]
#[1] 93.61 93.62 94.01 93.42 93.82 93.41 93.61 93.41 93.42 93.42 93.61 93.62 93.42 93.81

#[[4]]
#[1] 94.22 94.22 94.22 94.22 94.21

#[[5]]
#[1] 95.61 95.02 94.62 94.81

# Check the quality of the data fitting to the model
# p-val gives the risk error: the higher it is, the better the fitting is
ksfit = VARCORE_calculKSFit(clusC1, C1)
ksfit$pval
# [1] 0.205
# This means that the risk error to reject the hypothesis that the data is issued from a 
# Gaussian mixture model is 20.5%.   20.5% is not a small risk, so we do not reject 
# the assumption that the data C1 fits well the computed Gaussian mixture model clusC1.


# Plot fitting check
VARCORE_plotCDF_fit(clusC1,C1,maintitle="Estimated GM and empirical CDF", subtitle="Kolmogorov-Smirnov Fitting Test", xlabel="Data Values",ylabel="Probabilities")

# Demo2

# Load three sets of data, corresponding to three program versions: C1, C2 and C3 
C1 <- read.csv("ammp-C1", header=F)$V1
C2 <- read.csv("ammp-C2", header=F)$V1
C3 <- read.csv("ammp-C3", header=F)$V1

# Non-parametric performance comparison does not require clustering
# Estimate the probability that C2 < C3
p=VARCORE_probXY_nonparam(C2,C3)
p
#0.4989594 

# Estimate the probability that C1 < C3
p=VARCORE_probXYmannwhitney(C1,C3)
p
#1

# Estimate the probability that C2=min(C1,C2,C3)
p=VARCORE_probX1min_nonparam(list(C2,C1,C3))
p
#[1] 0

# Estimate the probability that C1=min(C1,C2,C3)
p=VARCORE_probX1min_nonparam(list(C1,C2, C3))
p
#[1] 1

# Compute the mean difference between C1 and C2
VARCORE_meandiff_nonparam(C1,C2)
#[1] 7.507419

# Parametric performance comparison
# Start by building Gaussian mixture model for each of C1, C2 and C3
clusC1 <- VARCORE_Clustering(C1)
clusC2 <- VARCORE_Clustering(C2)
clusC3 <- VARCORE_Clustering(C3)

# Estimate the probability that C2 < C3
p=VARCORE_probXY_Mclust(clusC2,clusC3)
p
#0.5092883 

# Estimate the probability that C1 < C3
p=VARCORE_probXY_Mclust(clusC1,clusC3)
p
#0.9999989

# Estimate the probability that C2=min(C1,C2,C3)
p=VARCORE_probX1min_Mclust(list(clusC2, clusC1, clusC3))
p
#[1]  2.891931e-07

# Estimate the probability that C1=min(C1,C2,C3)
p=VARCORE_probX1min_Mclust(list(clusC1, clusC2, clusC3))
p
#[1] 0.9999668
# Compute the mean difference between C1 and C2
val=VARCORE_meandiff_Mclust(clusC1,clusC2)
val
#[1] 7.507419

# Compute the alpha-quartile of C1, for instance alpha=0.33
val=VARCORE_quantile_Mclust(0.33, clusC1)
val
#[1] 85.58795

# Compute the probability that C1 < 85
p=VARCORE_probXa(clusC1, 85)
p
#[1] 0.1874038

# In this example, the first version of the code is certainly always faster than 
# the other two versions (and each of these two versions has about the same 
# chance of being lower than the other), and it is estimated that about 33% of 
# the time the first version has an execution time lower than 85.59 (and the 
# probability that the execution time is lower than 85 drops to about 19%)

