# This is a R file containing various functions of the VARCORE project
# Authors: Julien Worms (Université Vesailles Saint-Quentin) and Sid Touati (Université Nice Sophia Antipolis) 
# Copyright 2017
# This software can be distributed freely, but cannot be sold nor be part of a commercial software.

library(mclust)



######################################################
# Function VARCORE_Clustering: compute an estimated gaussian mixture model 
# based on a data sample X, by using the Mclust package (allowing for
# constant as well as distinct variances for the mixture components)
# Inputs: 
# - data: a sample containing data (a vector)
# - MaxClus: Maximal allowed number of clusters (in the search for the best gaussian mixture model, if NbClus below is 0)
# - NbClus: imposed number of clusters. If 0, let the function search for an adequate number of clusters
# Output: 
# - an object of type Mclust, which contains a clustering model (and information about other models tested)
######################################################
VARCORE_Clustering <- function(data, MaxClus=9, NbClus=0, MinClus=1){
  if(var(data) == 0){
    warning("VARCORE_Clustering: All data are identical. Clustering not possible !")
    return (FALSE)
  }
  if ( NbClus == 0 ) {     
    mclus <- Mclust(data,modelNames=c("E","V"),G=MinClus:MaxClus)
  } else {
    mclus <- Mclust(data,modelNames=c("E","V"),G=NbClus)
  }
 mclus 
}

##################
# Function VARCORE_calculKSFit  
# This function measures, using a Kolmogorov-Smirnov (KS) metric,the quality 
#      of fitting between data and the estimated gaussian mixture model. 
# Input:  
# - clus: a Mclust object (clustering result) , generally the GM model estimated from the data 
# - data: the data sample
# - nb : the chosen number of bootstrap replications (by default, 200)
# - coef : by default 1 ; if lower than 1, implies some sort of "subsampled bootstrap" (the bootstrap samples are not of size n, but of size coef*n)
# - trace: option, if yes, plot the empirical and estimated cumultative distribution function (CDF), and shows the point where the maximal gap between the two function occurs. 
# - calculpval: option TRUE by default, ask to compute the p-value of the fitting test, computed by Bootstrap/MonteCarlo simulation using VARCORE_simulmelange function
# - verbose: FALSE by default, otherwise outputs a little report on the screen
# 
# Output: a list that contains in order
# - the KS statistic ($statistic)
# - the p-value of the fitting test ($pval). The higher is better for data fitting.
# - the vector of KS distances produced by the bootstrap procedure ($KSdist)
################
VARCORE_calculKSFit <- function(clus, data, nb=200 , coef=1, trace=FALSE, calculpval=TRUE, verbose=FALSE , reeval=TRUE , param = TRUE){
  prm <- clus$parameters
  k  <- clus$G
  pichap <- as.vector(prm$pro) 
  muchap <- as.vector(prm$mean)
  if (clus$modelName == "E"){ 
    sigchap <- rep( sqrt(prm$variance$sigmasq) , k )
  } else {
    sigchap <- sqrt(prm$variance$sigmasq)
  } 
  n=length(data)
  
  Fchap <- rep(0,n)
  sorteddata <- sort(data)
  
  for (i in 1:n) {     # computation of the empirical cdf at the observations Xi
    Fchap[i] = sum ( pichap * pnorm(sorteddata[i],muchap,sigchap) ) 
  } 
  isurn = (1:n)/n  
  im1surn = isurn - 1/n
  m = matrix(c(isurn-Fchap,Fchap-im1surn),2,n,byrow=T)
  # computation of the Kolmorgorov-Smirnov statistic on the data  
  calculKSStat =  max( apply(m,2,"max") )
  
  # If asked for, a graph of the empirical and estimated GM cdf is performed
  
  if (trace == TRUE) { 
    VARCORE_plotCDF_fit(clus,data,maintitle="Estimated GM and empirical CDF", subtitle="Kolmogorov-Smirnov Fitting Test",xlabel="Data Values",ylabel="Probabilities")
    nummax = which.max( apply(m,2,"max") )
    points(sorteddata[nummax],nummax/n)
  }
  
  # If the computation of a p-value for the goodness-of-fit test was asked for, it is now
  # evaluated by parametric bootstrap simulation . This implies 
  #  - simulating nb samples following the estimated GM distribution (parametric bootstrap)
  # and then, for each such bootstrap sample,
  #  - computing the KS distance between the empirical distribution of this bootstrap sample
  #    and the gaussian mixture distribution re-estimated from it 
  # The p-value of the bootstrap-calibrated goodness-of-fit test is then equal to the
  # proportion of these distances which are greater than the KS distance of the original sample
  
  if (calculpval == TRUE) {
    
    KSsimules = rep(0,nb)  
    # will contain the KS statistics values which will be compared to the observed KS statistic 
    # calculKSstat which will allow evaluation of the p-value of the goodness-of-fit test
    no = floor(coef*n)
    # no will be the size of the bootstrap samples ; by default,
    # it is equal to n, but it may be chosen lower than n (when coef<1)
    # when some sort of "subsampled bootstrap" is decided
    
    FchapMC = rep(0,no)
    
    melangeMC = VARCORE_simulmelange(clus,no,nb)  
    # this matrix contains the nb bootstrap samples of size n 
    
    for (j in 1:nb) {
      sortedmelange = sort( melangeMC[,j] )
      
      # the GM model is now "re"-estimated on the basis of the bootstrap sample : this is from this 
      # GM model that the Kolm-Smirnov distance of the bootstrap empirical cdf will be computed. 
      # The possible numbers of clusters for the estimation of this new GM model are restricted
      # to five : the same as that for the original data, or 1 or 2 less or 1 or 2 more 
      
      clusB <- VARCORE_Clustering( sortedmelange , MinClus=max(1,k-2) , MaxClus=k+2 )
      
      prmB <- clusB$parameters
      kB  <- clusB$G
      pichapB <- as.vector(prmB$pro) 
      muchapB <- as.vector(prmB$mean)
      if (clusB$modelName == "E"){ 
        sigchapB <- rep( sqrt(prmB$variance$sigmasq) , kB )
      } else {
        sigchapB <- sqrt(prmB$variance$sigmasq)
      } 
      
      # The KM distance between the bootstrap sample and the re-evaluated GM model is computed 
      isurno = (1:no)/no   ;   im1surno = isurno - 1/no
      for (i in 1:no) { FchapMC[i] = sum ( pichapB * pnorm(sortedmelange[i],muchapB,sigchapB) ) }  
      KSsimules[j] = max ( max( isurno - FchapMC ) , max( FchapMC - im1surno ) )
    }
    
    # We have a sample of size nb of the distribution of the KS statistic under the null hypothesis :
    # -> we now count the proportion of those which are greater than the observed value KS_obs on the initial data
    pvaleur = mean( KSsimules > calculKSStat  ) 
    
    # The result can be displayed on the screen (if verbose=T)
    
    nummax = which.max( apply(m,2,"max") )
    if(verbose){
      cat("Result : Dn_obs=",calculKSStat,"\n ( obtained at ",nummax,"-th position among ",n,", i.e. at  x=",sorteddata[nummax],")\n")
      cat("p-value of the goodness-of-fit test, obtained by Bootstrap :",round(100*pvaleur,digits=1),"% \n")
    }
  }
  list(statistic = calculKSStat , pval = pvaleur , KSdist = KSsimules ) 
}

##################
# Function fonction : VARCORE_plotCDF_fit 
#  This function plots conjointly the empirical (cumulative) distribution function 
#  of the data and the (cumulative) distribution function of a previously computed
#  gaussian mixture model (generally, one obtained from the input datat)
# Input
# - clus:  a Mclust object (clustering result)
# - data: a data sample (vector)
# - maintitle: an overall title for the plot
# - subtitle: a sub title for the plot
# - label: a title for the x axis
# - ylabel: a title for the y axis
# - plot_legend: (FALSE by default) plot a legend
# Output 
# - no output, just a graph on the screen
############################
VARCORE_plotCDF_fit <- function(clus, data, maintitle="",subtitle="",xlabel="",ylabel="", plot_legend=FALSE) { 
  
  prm <- clus$parameters
  k  <- clus$G
  pichap <- as.vector(prm$pro) 
  muchap <- as.vector(prm$mean)
  if (clus$modelName == "E"){ sigchap <- rep( sqrt(prm$variance$sigmasq) , k )
  } else {  sigchap <- sqrt(prm$variance$sigmasq) }   

  etendue=max(data)-min(data)
  absc = seq(min(data)-0.05*etendue, max(data)+0.05*etendue, length.out=300 )
  
  Fchap = absc 
  for (i in 1:length(absc)) 
  { Fchap[i] = sum ( pichap * pnorm(absc[i],muchap,sigchap)  ) }
  
  ld = length(data) ; 
  plot(sort(data), cumsum(rep(1,ld))/ld,type='s',col='red',main=maintitle,sub=subtitle,xlab=xlabel,ylab=ylabel)
  lines(absc,Fchap,type='l')
  if (plot_legend) legend("bottomright", legend = c("CDF (theoretical)", "CDF (empirical)"), col = c("black",  "red"), lwd=c(1,1), lty=c(1,1))
  
}

##################
# Function fonction : VARCORE_plot_clustering_result 
#  This function plots conjointtly the histogram of the data and the density 
#  function of a previously computed clustering result (generally obtained from the input data)
# Input
# - clus:  a Mclust object (clustering result)
# - data: a data sample
# - maintitle: an overall title for the plot
# - subtitle: a sub title for the plot
# - label: a title for the x axis
# - ylabel: a title for the y axis
# - plot_legend: (FALSE by default) plot a legend
# - plot_sdensity: (TRUE by default) plot the density function of the sample data (blue)
# - plot_tdensity: (TRUE by default) plot the density function of the gaussian mixture (red)
# - plot_rug: (FALSE by default) makes the original raw data appear on the x-axis
# Output 
# - no output, just a graph on the screen
############################
VARCORE_plot_clustering_result  <- function(clus, data, maintitle="",subtitle="",xlabel="",ylabel="", plot_sdensity=TRUE, plot_tdensity=TRUE, plot_legend=FALSE, plot_rug=FALSE) { 
  
  prm <- clus$parameters
  k  <- clus$G
  pichap <- as.vector(prm$pro) 
  muchap <- as.vector(prm$mean)
  if (clus$modelName == "E"){ sigchap <- rep( sqrt(prm$variance$sigmasq) , k )
  } else {  sigchap <- sqrt(prm$variance$sigmasq) }   
  
  etendue=max(data)-min(data)
  absc = seq(min(data)-0.05*etendue, max(data)+0.05*etendue, length.out=300 )
  
  fchap = absc 
  for (i in 1:length(absc)) 
  { fchap[i] = sum ( pichap * dnorm(absc[i],muchap,sigchap)  ) }
  
  ld = length(data) 
  
  hist(data,main=maintitle,sub=subtitle,xlab=xlabel,ylab=ylabel)
  
  if(plot_sdensity) {
    par(new=TRUE)
    plot(density(data),col='blue', axes=FALSE,main="",sub="",xlab="",ylab="")
  }
  
  if(plot_tdensity) {
    par(new=TRUE)
    plot(absc, fchap,type='l',col='red',axes=FALSE, main="",sub="",xlab="",ylab="")
  }
  if(plot_rug)  {
    par(new=TRUE)
    rug(data)
  }
  
  if (plot_legend) {
    par(new=TRUE)
    legend("topright", legend = c("histogram of the data", "estimated density (kernel)", "estimated density (mixture)"), col = c("black", "blue", "red"), lwd=c(1,1,1), lty=c(1,1,1), cex=0.6)
  }
}

###############
# Function VARCORE_simulmelange 
# This functions simulates N samples of gaussian mixtures, each sample is of size n
# Input:
# - clus: a Mclust object (clustering result) used to simulate N samples
# - n: number of simulated values per sample
# - N : number of samples
# Output: 
# - a matrix N x n containing realizations of the gaussian mixture distribution given as input
#   (or a vector of length n if N=1)
################
VARCORE_simulmelange <- function(clus,n,N){
  simulmelange <- matrix(rep(0,n*N),n,N)  
  for (jj in 1:N) {
    # Use of the sim function from the Mclust package
    simulmelange[,jj] <- sim(clus$modelName,clus$parameters,n)[,2] 
  }
  if (N == 1) { simulmelange = as.vector(simulmelange) } 
  simulmelange
}

##########################
# Function VARCORE_generevaleursmelangegaussiennes
# This function generates a sample of n data from a gaussian mixture given as input 
# through explicit vectors (and not as a Mclust object as in VARCORE_simulmelange).
# Input
# - K : number of gaussian components of the mixture
# - PiK: vector of K values, which are the weights of the gaussian components
# - MuK:  vector of K values, which are ontaining the means of the gaussian components
# - SigmaK: vector of K values which are the standard deviations of the gaussian components
# Output: 
# - a vector of n values issued from the gaussian mixture distribution given as input
##########################
VARCORE_generevaleursmelangegaussiennes <- function(n, K, PiK, MuK, SigmaK  ){
  
  if (length(PiK) != K){
    cat("The length of PiK vector is different than", K)
    return (FALSE)
  }
  if (length(MuK) != K){
    cat("The length of MuK vector is different than", K)
    return (FALSE)
  }
  if (length(SigmaK) != K){
    cat("The length of SigmaK vector is different than", K)
    return (FALSE)
  }
  sumPiKm1 = c(0,cumsum(PiK[-K]))
  U = runif(n)
  Y = rep(0,n)  # Y will contain the clusters numbers from which the data are issued
  echantillon = rep(0,n)
  for (i in 1:n) { Y[i] = sum ( U[i] > sumPiKm1 ) }
  for (i in 1:n) { echantillon[i] = rnorm(1,MuK[Y[i]],SigmaK[Y[i]]) }
  
  generevaleursmelangegaussiennes = echantillon
}



####################################
# Estimating and testing functions #
####################################


####################
# Function VARCORE_check_normality: checks if a sample of data is drawn from a gaussian function
# Input:
#   - X : a data sample
#   - alpha: desired risk level (5% by default)
# Ouput: 
# - FALSE if evidence is found that the sample X is not issued from a gaussian distribution, TRUE otherwise
###########################
VARCORE_check_normality <- function(X, alpha=0.05) { # alpha is the risk of error = 1-conflevel 
  #lili <-  lillie.test(X);
  #if(is.na(lili$p.value) || is.infinite(lili$p.value)) lili$p.value <- 0;
  if (var(X)==0){
    return (FALSE); # all values are identical
  }
  shap <- shapiro.test(X);
  if(is.na(shap$p.value) || is.infinite(shap$p.value)) shap$p.value <- 0;
  #ad <- ad.test(X)
  #if(is.na(ad$p.value) || is.infinite(ad$p.value)) ad$p.value <- 0;
  #cvm <- cvm.test(X) Fonction bizzare ?<80>? v?<88>rifier
  #sf <- sf.test(X)
  #if(is.na(sf$p.value) || is.infinite(sf$p.value)) sf$p.value <- 0;
  #pval <- max(lili$p.value, shap$p.value, ad$p.value,  sf$p.value);
  pval <- shap$p.value
  if (pval <= alpha) {
    return (FALSE)
  }
  else { # if we do not succeed in rejecting the normality with risk alpha, we admit (without proof) that the distribution is normal. But this is not necessarily true !
    return (TRUE)
  }
}

#############################
# Function : VARCORE_probXY_Mclust
# Computes P(X<Y) when X and Y are two random variables following (two) gaussian mixture 
# distributions which have been derived by clustering of data samples.
# This is thus a parametric estimation of the true probability P(X<Y) when X and Y
# follow the (unknown) underlying distributions of the data samples.
# Input:
# - clus1: a Mclust object (clustering result for the "X sample") 
# - clus2: a Mclust object (clustering result for the "Y sample") 
# - Delta: optional, by default=0, otherwise allows for the computation of P(X<Y+Delta)
# Output: 
# - the (estimated) probability that X < Y (or Y+Delta)
#############################
VARCORE_probXY_Mclust <- function(clus1,clus2,Delta=0) {

  k1 <- clus1$G
  pichap1  <- as.vector(clus1$parameters$pro)
  muchap1  <- as.vector(clus1$parameters$mean)
  if (clus1$modelName == "E") {
    sigchap1 <-  rep( sqrt(clus1$parameters$variance$sigmasq) , k1 )  # vector of std deviations
  }else {
    sigchap1 <-  sqrt(as.vector(clus1$parameters$variance$sigmasq)) # vector of std deviation
  }
  
  k2 <- clus2$G
  pichap2  <- as.vector(clus2$parameters$pro)
  muchap2  <- as.vector(clus2$parameters$mean)
  if (clus2$modelName == "E") {
    sigchap2 <-  rep( sqrt(clus2$parameters$variance$sigmasq) , k2 )  # vector of std deviations
  }else {
    sigchap2 <-  sqrt(as.vector(clus2$parameters$variance$sigmasq)) # vector of std deviation
  }
  
  sommetemp = rep(0,k1)
  for (k in 1:k1) {
    sommetemp[k] = sum ( pichap2 * pnorm( (muchap2 + Delta  - muchap1[k]) / sqrt( (sigchap1[k])^2 + sigchap2^2 ) )   )
  }
  probXY <- sum( pichap1 * sommetemp )  
} 

#############################
# Function : VARCORE_probXY
# Compute P(X<Y) when X and Y are two random variables following a gaussian mixture models
# (these models are now given explicitely as input, not in the Mclust format)
# Input:
# - PiKX: vector of K values, which are the weights of the gaussian components for X
# - MuKX:  vector of K values, which are ontaining the means of the gaussian components for X
# - SigmaKX: vector of K values which are the standard deviations of the gaussian components for X
# - PiKY: vector of K values, which are the weights of the gaussian components for Y
# - MuKY:  vector of K values, which are ontaining the means of the gaussian components for Y
# - SigmaKY: vector of K values which are the standard deviations of the gaussian components for Y
# - Delta (= 0 by default): optional. Enables computation of P(X<Y+Delta) instead of P(X<Y)
# Output: 
# - the probability that X < Y (or Y+Delta)
#############################
VARCORE_probXY <- function(PiKX,MuKX,SigmaKX, PiKY,MuKY , SigmaKY, Delta=0) {
  nX=length(PiKX)
  if(length(MuKX) != nX || length(SigmaKX) != nX){
    cat("Error in VARCORE_probXY. The size of PiK, MuK, SigmaK must be the same for X:", nX)
    return (-1)
  }
  nY=length(PiKY)
  if(length(MuKY) != nY || length(SigmaKY) != nY){
    cat("Error in VARCORE_probXY. The size of PiK, MuY, SigmaK must be the same for Y:", nY)
    return (-1)
  }
  sommetemp = rep(0,nX)
  for (k in 1:nX) {
    sommetemp[k] = sum ( PiKY * pnorm( (MuKY + Delta  - MuKX[k]) / sqrt( (SigmaKX[k])^2 + SigmaKY^2 ) )   )
  }
  probXY <- sum( PiKX * sommetemp )  
}

#############################
# Function : VARCORE_probXY_nonparam 
# Estimates nonparametrically P(X<Y) when X and Y are two data samples, 
# using the Mann-Whitney statistic
# Input:
# - X : a data sample
# - Y : a data sample
# - Delta (= 0 by default): optional. Enables computation of P(X<Y+Delta) instead of P(X<Y)
# Output: 
# - the (estimated) probability that X < Y (or Y+Delta)
#############################
VARCORE_probXY_nonparam  <- function(X, Y, Delta=0) {
  nY <- length(Y)
  nX <- length(X)
  probXY <- (( wilcox.test(Y+Delta,X)$statistic ) / (nX*nY))
}

#############################
# Function : VARCORE_probXa
# Compute P(X<a) when X is a random variable following a model of gaussian mixture distribution
# Input:
# - clus: a Mclust object (clustering result) , describing the distribution of X
# - a : a real value
# Output: 
# - the (estimated) probability that X < a
#############################
VARCORE_probXa <- function(clus,a) {
  K <- clus$G
  pichap  <- as.vector(clus$parameters$pro)
  muchap  <- as.vector(clus$parameters$mean)
  if (clus$modelName == "E") {
    sigchap <-  rep( sqrt(clus$parameters$variance$sigmasq) , K )  # vector of std deviations
  }else {
    sigchap <-  sqrt(as.vector(clus$parameters$variance$sigmasq)) # vector of std deviation
  }
 
  sommep = 0
  for (k in 1:K) {
    sommep = sommep + ( pichap[k] * pnorm( a, mean=muchap[k], sd=  sigchap[k])  )
  }
  sommep  
} 

#############################
# Function VARCORE_quantile_Mclust
# This function computes the quantile of order alpha of a gaussian mixture distribution
# Input: 
# - alpha : the desired order 0 < alpha < 1 for the quantile 
# - clus: a MClust object (clustering result) 
# Output: 
# - the (estimated) quantile 
#
# Examples:
# the median is estimated by VARCORE_quantile_Mclust(0.5, clus)
# the first quartile Q1 is estimated by VARCORE_quantile_Mclust(0.25, clus)
# the third quartile Q3 is estimated by VARCORE_quantile_Mclust(0.75, clus)
#################
VARCORE_quantile_Mclust  <-  function( alpha , clus ){
  k <- clus$G
  pichap  <- as.vector(clus$parameters$pro)
  muchap  <- as.vector(clus$parameters$mean)
  if (clus$modelName == "E") {
    sigchap <-  rep( sqrt(clus$parameters$variance$sigmasq) , k )  # vector of std deviations
  }else {
    sigchap <-  sqrt(as.vector(clus$parameters$variance$sigmasq)) # vector of std deviation
  }
  
  
  FdRmelangemoinsalpha <- function(x){
    sum( pichap * pnorm( (x-muchap)/sigchap ) ) - alpha
  } 
  bornes = c( min(muchap-3*sigchap),max(muchap+3*sigchap)  )
  quantilemelange <- uniroot( FdRmelangemoinsalpha , bornes )$root   
}

#################################################
# Function VARCORE_quantile
# Computes the quantile of order alpha of a gaussian mixture distribution
#   (this mixture model is now given explicitely as input, not in the Mclust format)
# Input
# - alpha: the desired order (alpha=0.5 for the median). 0 < alpha < 1 
# - PiK: vector of values, which are the weights of the  gaussian mixture components
# - MuK:  vector of values, which contains the means of the gaussian mixture components
# - SigmaK: vector of values, which are the standard deviations the gaussian mixture components
# Output: 
# - the (estimated) quantile of ordre alpha
# Example: 
#    VARCORE_quantile(0.5, clus) computes the median of the gaussian mixture given as input (PiK, MuK, SigmaK)
#################################################
VARCORE_quantile <-  function(alpha,  PiK, MuK, SigmaK){
  n <- length(PiK)
  if(length(MuK) != n || length(SigmaK) != n){
    cat("Error in VARCORE_quantile. The size of PiK, MuK, SigmaK must be the same:", n)
    return (-1)
  }
  FdRmelangemoinsalpha <- function(x){
    sum( PiK * pnorm( (x-MuK)/SigmaK ) ) - alpha
  } 
  bornes = c( min(MuK-3*SigmaK),max(MuK+3*SigmaK)  )
  quantilemelange <- uniroot( FdRmelangemoinsalpha , bornes )$root   
}


#############################
# Function : VARCORE_probX1min_Mclust
# Computes P(X1<=min(X2,..,XK)) when X1,X2,..,XK are K random variables 
#  following (generally different) gaussian mixture models derived from clustering
# Input:
# - listclus: a list of K objects of type Mclust (clustering results) 
# Output: 
# - the estimated probability that X1 <= min(X2,..,XK) , estimated via the parametric approach
#
# Note : needs the auxiliary function "VARCORE_tempfunction" defined below, 
#  (itself needing the auxiliary functions "VARCORE_densityX1" and "VARCORE_CdFofminX2Xp")
#############################
VARCORE_probX1min_Mclust <- function(listclus){
  
  # Preliminaries
  ###############
  K <- length(listclus)
  # K is the number of models, ie the number of samples under study
  kchap   <- vector("integer",K)
  pichap  <- vector("list",K)
  muchap  <- vector("list",K)
  sigchap <- vector("list",K)
  # the lists above will contain the characteristics of the gaussian mixtures models given as input,
  # in the form of vectors (which lengths will be contained in the vector kchap)
  lowerbound <- vector("integer",K)
  upperbound <- vector("integer",K)
  # the vectors above contain the bounds needed for the integration providing the computation of the probability
  for (j in 1:K) {
    # Write in convenient lists/vectors the characteristics of the input gaussian mixtures models 
    clusj <- listclus[[j]] 
    kchap[j]  <- clusj$G
    pichap[[j]]  <- as.vector(clusj$parameters$pro)
    muchap[[j]]  <- as.vector( clusj$parameters$mean )
    if (clusj$modelName == "E") {
      sigchap[[j]] <-  rep( sqrt(clusj$parameters$variance$sigmasq) , kchap[j] )  # vector of std deviations
    }else {
      sigchap[[j]] <-  sqrt(as.vector(clusj$parameters$variance$sigmasq)) # vector of std deviation
    }
    lowerbound[j] = min( muchap[[j]]-4*sigchap[[j]] ) 
    upperbound[j] = max( muchap[[j]]+4*sigchap[[j]] )
  }
  
  # Numerical evaluation of P(X1<=min(X2,..,Xp)) (as an integral)
  #############################################
  
  integral = integrate( VARCORE_tempfunction , min(lowerbound) , max(upperbound) , pic=pichap , muc=muchap , sigc=sigchap , p=K)
  
  probX1isminimum <- integral$"value"
  
}

#############################
# Function : VARCORE_probX1min_nonparam
# Given K samples (of possibly different sizes), this function estimates P(X1<=min(X2,..,XK)) 
# when X1,X2,..,XK are the K random variables from which the K samples are drawn.
# This estimation is non-parametric, i.e. it does NOT assume that these samples
# are drawn from a particular family of distributions (eg: gaussian mixtures)
#
# Caution : when the samples are not of small size, the computation time can be quite lengthy
#
# Input:
# - datalist : a list containing a number of data samples (of possibly different sizes)
# Output: 
# - the (nonparametrically) estimated probability that X1 <= min(X2,..,XK) 
#
# Note on the method : 
#   The data are pooled in one big sample, and the numbers of the samples they come from
#   are stored in a vector "indices". The pooled data are sorted, and then the function
#   goes through it in ascending order : each time one of the data from the first sample (X1),
#   say x, is met, the number and indices of all the data values (from the other samples) 
#   that are greater or equal to x, is appropriately stored. When all the pooled data have
#   been gone through, the non-parametric estimator of P(X1<=min(X2,..,XK)) is then equal
#   to the sum of the products of these numbers ; ties have to be carefully taken into account.
##############################
VARCORE_probX1min_nonparam <- function(datalist){
 # Number of data samples
 k <- length(datalist)
 # Construction of the pooled sample and the corresponding indices vector, then sort
 for (j in 1:k){
    if (j == 1){
      Z <- datalist[[1]]
      indices <- rep(1,length(Z)) 
      eff <- rep(0,k-1)
    } else {
      x <- datalist[[j]]  ;  nj <- length(x)
      Z <- c(Z,x)
      indices <- c(indices,rep(j,nj))
      eff[j-1] <- nj
    }
  }
  S <- sort(Z,index.return=TRUE)
  Z <- S$x
  indices <- indices[S$ix]
  nbtot <- length(Z)
  # At this time, the vector eff contains the total numbers of data in all samples except the first.  
  #
  # In order to have the method work correctly, it is required that, in the pooled sample, 
  # if ties are present, then in the tied data those coming from the first sample must come first.
  # Note that this may be unneeded, since the sort function seems to do the job already.
  m=1
  z=Z[1]
  nequal=1
  while (m<nbtot){
   m=m+1
   z2=Z[m]
   if (z2 == z){
    nequal=nequal+1
   } else { # as soon as a set of tied data is "passed", we sort their indices (ascending)
   indi <- indices[(m-nequal):(m-1)]
   #cat("z=",z,",z2=",z2,",nequal=",nequal,",indi=",indi,"\n")
   indi <- sort(indi)
   #cat("... puis indi=",indi,"\n")
   indices[(m-nequal):(m-1)] <- indi  
   z <- z2
   nequal=1
   }
  }
  
  # The total number of K-tuples is computed now, before the vector eff gets modified
  nbtuples_tot <- prod(eff)*length(datalist[[1]])
  
  # Now the data is gone through a second time, in order to recover the needed information (in the matrix mateff)
  mateff = 0 
  for (m in 1:nbtot){
    num <- indices[m]
    if (num == 1){ # if a data value from the first sample is met, the number 
      # of data values which are equal or greater than it, is stored (in mateff)
      if ( is.matrix(mateff) == F ){ # initialization of mateff if it is the first X1 data to be encountered
        mateff = matrix(eff,nrow=1) 
      } else {  # otherwise the information is stacked below mateff
        mateff <- rbind( mateff , eff )   
      }
    } else { # if it isn't a data from the first sample, the vector eff is modified to indicate 
             # that a data has been exceeded (it won't be greater than any further data from the first sample) 
      eff[num-1] <- eff[num-1]-1   # a the very end, the vector eff will contain only zeros
    }  
  }  
  # The i-th line of the matrix mateff contains the numbers of data from samples 2 to k
  # which are greater or equal than the i-th data from the first sample.
  # It then suffices to take the sum of the products of its lines, and divide
  # by the product of the sample sizes, in order to estimate P(X1<=min(X2,..,Xk))
  
  probX1minNP <- sum( apply(mateff,1,prod) ) / nbtuples_tot
  probX1minNP
} 

#############################
# Function : VARCORE_meandiff_Mclust
# Compute E(|X-Y|) when X and Y are two random variables following gaussian mixture distributions 
# computed by Mclust clustering (from data), therefore enabling estimation of E(|X-Y|) from data  
# Input:
# - clus1: a Mclust object (clustering result), describing the distribution of X
# - clus2: a Mclust object (clustering result), describing the distribution of Y
# Output: 
# - the (estimated) expectation of the absolute difference |X-Y|
#############################
VARCORE_meandiff_Mclust <- function(clus1,clus2) {
  # recovering of the parameters of the two mixture distributions 
  
  k1 <- clus1$G
  pichap1  <- as.vector(clus1$parameters$pro)
  muchap1  <- as.vector(clus1$parameters$mean)
  if (clus1$modelName == "E") {
    sigchap1 <-  rep( sqrt(clus1$parameters$variance$sigmasq) , k1 )  # vector of std deviations
  }else {
    sigchap1 <-  sqrt(as.vector(clus1$parameters$variance$sigmasq)) # vector of std deviation
  }
  
  k2 <- clus2$G
  pichap2  <- as.vector(clus2$parameters$pro)
  muchap2  <- as.vector(clus2$parameters$mean)
  if (clus2$modelName == "E") {
    sigchap2 <-  rep( sqrt(clus2$parameters$variance$sigmasq) , k2 )  # vector of std deviations
  }else {
    sigchap2 <-  sqrt(as.vector(clus2$parameters$variance$sigmasq)) # vector of std deviation
  }
  
  
  temp <- rep(0,k1)
  for (k in 1:k1) {
    temp1 = (muchap1[k]-muchap2)*(1-2*pnorm( (muchap2-muchap1[k])/sqrt(sigchap1[k]^2+sigchap2^2) ) )
    temp2 = sqrt( 2*(sigchap1[k]^2+sigchap2^2)/pi ) * exp( -(muchap1[k]-muchap2)^2 / (2*(sigchap1[k]^2+sigchap2^2)) ) 
    temp[k]  = sum ( pichap2 * (temp1 + temp2) )  
  }
  meandiff <- sum( pichap1 * temp )  
} 

# Computes same thing than VARCORE_meandiff_Mclust, except that the parameters are not Mclust objects
VARCORE_meandiff_param <- function(PiKX,MuKX,SigmaKX, PiKY,MuKY , SigmaKY) {
  # recovering of the parameters of the two mixture distributions 
  
  k1 <- length(PiKX)
  pichap1  <- PiKX
  muchap1  <- MuKX
  sigchap1 <- SigmaKX
  k2 <- length(PiKY)
  pichap2  <- PiKY
  muchap2  <- MuKY
  sigchap2 <-  SigmaKY
  
  
  temp <- rep(0,k1)
  for (k in 1:k1) {
    temp1 = (muchap1[k]-muchap2)*(1-2*pnorm( (muchap2-muchap1[k])/sqrt(sigchap1[k]^2+sigchap2^2) ) )
    temp2 = sqrt( 2*(sigchap1[k]^2+sigchap2^2)/pi ) * exp( -(muchap1[k]-muchap2)^2 / (2*(sigchap1[k]^2+sigchap2^2)) ) 
    temp[k]  = sum ( pichap2 * (temp1 + temp2) )  
  }
  meandiff <- sum( pichap1 * temp )  
} 


#############################
# Function : VARCORE_meandiff_nonparam
# Estimates nonparametrically E(|X-Y|) from two samples which respective 
#    distributions are the same as that of random variables X and Y
# Input:
# - X : data 
# - Y : data 
# Output: 
# - the nonparametric estimation of the expectation of the absolute difference |X-Y|
#############################
VARCORE_meandiff_nonparam <- function(X,Y) {
  nX <- length(X)
  tempmean <- rep(0,nX)
  for (i in 1:nX) { tempmean[i] = mean( abs( X[i]-Y ) ) }
  mean(tempmean) 
}

#################################################
# Function VARCORE_nbmodes_estimation: 
# This function evaluates the number of modes (local maxima) of the 
# gaussian mixture probability distribution provided in input as a Mclust object
# Input: 
# - a Mclust object (clustering result)
# Output: 
# - an integer variable containing the evaluated number of modes of the estimated GM pdf
#################################################
VARCORE_nbmodes_estimation <- function(clus){
  k <- clus$G
  pichap  <- as.vector(clus$parameters$pro)
  muchap  <- as.vector(clus$parameters$mean)
  if (clus$modelName == "E") {
    sigchap <-  rep( sqrt(clus$parameters$variance$sigmasq) , k )  # vector of std deviations
  }else {
    sigchap <-  sqrt(as.vector(clus$parameters$variance$sigmasq)) # vector of std deviation
  }
  precision <- min(sigchap/2) 
  xmin <- min(muchap)-3*min(sigchap) ; xmax <- max(muchap)+3*max(sigchap)
  xgrid <- seq(xmin,xmax,by=precision)
  nbmodes <- 0
  signe <- 1
  for (i in 1:length(xgrid)){
    x <- xgrid[i] 
    derivpdf <- sum( - pichap * (x-muchap) * exp( -(x-muchap)^2/(2*sigchap^2) ) / (sigchap^3 * sqrt(2* pi))  ) 
    if ( sign(signe*derivpdf) == -1 ) { 
      signe <- -signe  
      if (signe == -1) {nbmodes <- nbmodes+1}
    }
  }
  nbmodes 
}

#################################################
# Function VARCORE_nbmodes_computation: 
# This function evaluates the number of modes (local maxima) of an underlying "true" 
# gaussian mixture probability distribution provided in input (but not in Mclust form) 
# Input: 
# - muk : a vector containing the values of the components means
# - sigmak : a vector containing the values of the components standard deviations
# - pik : a vector containing the values of the components weights
# (all these vectors must have the same length)
# Output: 
# - an integer variable containing the evaluated number of modes of the input GM pdf
#################################################
VARCORE_nbmodes_computation <- function(pik,muk,sigmak){
  precision <- min(sigmak/2) 
  xmin <- min(muk)-3*min(sigmak) ; xmax <- max(muk)+3*max(sigmak)
  xgrid <- seq(xmin,xmax,by=precision)
  nbmodes <- 0
  signe <- 1
  for (i in 1:length(xgrid)){
    x <- xgrid[i] 
    derivpdf <- sum( - pik * (x-muk) * exp( -(x-muk)^2/(2*sigmak^2) ) / (sigmak^3 * sqrt(2* pi))  ) 
    if ( sign(signe*derivpdf) == -1 ) { 
      signe <- -signe  
      if (signe == -1) {nbmodes <- nbmodes+1}
    }
  }
  nbmodes 
}


#################################################
# Function VARCORE_extract_cluster_information: 
# This function extracts the most important parameters from a Mclust object (gaussian mixture)
# Input: 
# - a Mclust object (clustering result)
# Output: 
# a data frame that contains
# - weights: vector  that contains the weights of the gaussian components
# - means: vector nbclus that contains the means of the gaussian components
#-  stdevs: vector  nbclus that contains the standard deviations of the gaussian components
# Of course, the three vectors (weights, means, stdevs) are of the same size, and equal to K, 
# the number of components of the gaussian mixtures
# Example
#  result$means[i] contains the the mean of the i th gaussian component.
#  length(result$means) is the total number of gaussian components
#################################################
VARCORE_extract_cluster_information <- function(clus){
  k <- clus$G
  pichap  <- as.vector(clus$parameters$pro)
  muchap  <- as.vector(clus$parameters$mean)
  
   if (clus$modelName == "E") {
    sigchap <-  rep( sqrt(clus$parameters$variance$sigmasq) , k )  # vector of std deviations
  }else {
    sigchap <-  sqrt(as.vector(clus$parameters$variance$sigmasq)) # vector of std deviation
  }
  
  result<- data.frame(weights=pichap, means=muchap, stdevs=sigchap)
}

#################################################
# Function VARCORE_extract_cluster_list : 
# This function extracts the list of clusters : in other words, it assigns to each data value
# one of the clusters of the gaussian mixture (the one from which it is the closest), and
# it gathers in a list the vectors of values which share the same cluster
# Input: 
# - clus: a Mclust object (clustering result)
# - data: a data sample (generally clus is the mixture estimated from X)
# Output: 
# - the list of data sorted by cluster
# Example: 
#    list <- VARCORE_extract_cluster_list(clus, data) 
#    list[[k]] contains all the data values which have been assigned to cluster number k  ; equals numeric(0) if none has been affected to the k-th cluster
#################################################
VARCORE_extract_cluster_list  <- function(clus, data){
  affectations <- predict(clus,data)$classification
  nbclust <- clus$G
  for (kk in 1:nbclust) {
    if (kk == 1) { listeclusters = list( data[affectations==1] ) } 
    else {  
      listeclusters = c( listeclusters , list(data[affectations==kk] ))
    }
  }
  listeclusters
}


#################################################
# Function VARCORE_mean_Mclust : compute the mean of a gaussian mixture given as parameter
# Input: 
# - clus a Mclust object (clustering result)
# Output: computed mean
# Example: 
#    VARCORE_mean_Clust(clus) computes the mean of the gaussian mixture clus
#################################################
VARCORE_mean_Mclust <-  function(clus){
  pichap  <- as.vector(clus$parameters$pro)
  muchap  <- as.vector(clus$parameters$mean)
  result <- pichap %*% muchap # scalar product
  result[1]
}

############################################################
# AUXILIARY FUNCTIONS (used in VARCORE_probX1min_Mclust above)
# The three functions below are not intended to be used 
#     outside the function VARCORE_probX1min_Mclust
###########################################################

VARCORE_densityX1 <- function(x,pic,muc,sigc){ 
  # this function computes the value of the density of X1 (which gaussian mixture distribution
  # is given as input) at every element of the real vector x. 
  # Input :
  # - a vector x of real numbers at which the density is evaluated
  # - the lists pic,muc,sigc (of size p) containing the characteristics of the p gaussian mixtures 
  # Output :
  # - the vectors of values of the density of X1 at every element of x
  resu=rep(0,length(x))
  for (l in 1:length(x)) {
    resu[l] = sum( pic[[1]]*dnorm( x[l] , muc[[1]] , sigc[[1]] ) ) 
  }
  resu
}

VARCORE_CdFofminX2Xp <- function(x,pic,muc,sigc,p) {
  # this technical function computes the value of the cumulative distribution function
  # of min(X2,..,Xp) at every element of x, where X2,..,Xp are distributed
  # as gaussian mixtures depicted in the various input parameters. 
  # Input :
  # - a vector x of real numbers at which the Cdf is evaluated
  # - the lists of vectors containing the characteristics of the gaussian mixtures
  # - the value of p
  # Output :
  # - the vector of values of P(min(X2,..,Xp)>s) (where s spans x)
  resu=rep(0,length(x))
  for (l in 1:length(x)) {
    tempor = rep(0,p)
    for (ll in 2:p) {
      tempor[ll] = 1 - sum( pic[[ll]]*pnorm( x[l] , muc[[ll]] , sigc[[ll]] ) ) 
      # tempor[ll] contains the value P(Xll>x[l])
    }
    resu[l] = prod( tempor[2:p] )  
    # resu[l] contains the value P(min(X2,..,Xp)>x[l]), as desired
  }
  resu
} 

VARCORE_tempfunction <- function(x,pic,muc,sigc,p){
  # this technical function has as integral the probability that X1<min(X2,..,Xp) ;
  # it is meant to be integrated at the end of the code VARCORE_probX1min_Mclust
  # Input : 
  # - a vector x of arbitrary length
  # Output :
  # - the vector value of (1-G)*f1 at every element of x 
  #       (where G is the CdF of min(X2,..,Xp) and f1 is the density of X1)
  VARCORE_CdFofminX2Xp(x,pic,muc,sigc,p) * VARCORE_densityX1(x,pic,muc,sigc) 
}

#Patch mclust provided by Luca Scrucca in november 2015, maintainer of Mclust
#----------------------------------------------------------------------------
qclass <- function (x, k) 
{
  x <- as.vector(x)
  # eps <- sqrt(.Machine$double.eps) 
  # numerical accuracy problem if scale of x is large, so make tolerance
  # scale dependent
  eps <- sd(x)*sqrt(.Machine$double.eps)
  q <- NA
  n <- k
  while(length(q) < (k+1))
  { n <- n + 1
    q <- unique(quantile(x, seq(from = 0, to = 1, length = n))) 
  }
  if(length(q) > (k+1))
  { dq <- diff(q)
    nr <- length(q)-k-1
    q <- q[-order(dq)[1:nr]]
  }
  q[1] <- min(x) - eps
  q[length(q)] <- max(x) + eps
  cl <- rep(0, length(x))
  for(i in 1:k) 
  { cl[ x >= q[i] & x < q[i+1] ] <- i }
  return(cl)
}
require(R.utils)
reassignInPackage("qclass", pkgName = "mclust", qclass)

# Simple example of usage of VARCORE functions
########################################
#data <- read.csv("datafile.csv", header=F)$V1
#clusX <- VARCORE_Clustering(data)
#VARCORE_plot_clustering_result(clusX,data, maintitle="performances model", subtitle="Gaussian Mixture Model",xlab="Execution Times",ylab="Probability/Frequency")
# rug(data)
